#ifndef SET_HPP_
#define SET_HPP_

#include <iterator>
#include <limits>

#include "support/concepts.hpp"

#include "Tree.hpp"

template<typename T>
class Set;

template<typename T>
class SetIterator;

template<typename T>
struct SetIterator
{
  friend class Set<T>;

  using value_type = T const;
  using reference = value_type&;
  using pointer = value_type*;
  using difference_type = std::ptrdiff_t;
  using iterator_category = std::bidirectional_iterator_tag;
  using node_type = TreeNode<T>;

  constexpr explicit SetIterator (node_type* n) noexcept : m_node (n)
  {
  }
  constexpr reference operator* () noexcept
  {
    return m_node->data;
  }
  constexpr pointer operator-> () noexcept
  {
    return &(m_node->data);
  }
  constexpr SetIterator&
  operator++ () noexcept
  {
    // TODO -- hint: find in-order successor
    return *this;
  }
  constexpr SetIterator
  operator++ (int) noexcept
  {
    SetIterator copy (*this);
    ++(*this);
    return copy;
  }
  constexpr SetIterator&
  operator-- () noexcept
  {
    // TODO -- hint: find in-order predecessor
    return *this;
  }
  constexpr SetIterator
  operator-- (int) noexcept
  {
    SetIterator copy (*this);
    ++(*this);
    return copy;
  }
  friend constexpr bool
  operator== (SetIterator i, SetIterator j) noexcept
  {
    // TODO -- compare node pointers for equality
    return true;
  }
  friend constexpr bool
  operator!= (SetIterator i, SetIterator j) noexcept
  {
    return !(i == j);
  }

private:
  node_type* m_node;
};

template<typename T>
class Set
{
  using node_type = TreeNode<T>;

public:
  using value_type = T;
  using size_type = std::size_t;
  using difference_type = std::ptrdiff_t;
  using reference = value_type&;
  using const_reference = value_type const&;
  using pointer = value_type*;
  using const_pointer = value_type const*;
  using iterator = SetIterator<T>;
  using const_iterator = iterator;
  using reverse_iterator = std::reverse_iterator<iterator>;
  using const_reverse_iterator = std::reverse_iterator<const_iterator>;

private:
public:
  Set () = default;
  Set (Set const&) = default;
  Set (Set&&) = default;

  ~Set () = default;

  Set&
  operator= (Set const&) = default;

  Set&
  operator= (Set&&) = default;

  template<typename InputIterator,
           Requires (concepts::ForwardIterator, InputIterator)>
  Set (InputIterator first, InputIterator last)
  {
    insert (first, last);
  }

  Set (std::initializer_list<T> ilist) : Set (ilist.begin (), ilist.end ())
  {
  }

  iterator
  begin () noexcept
  {
    // TODO -- return iterator that points to the minimum node in the tree
    return iterator (nullptr);
  }

  const_iterator
  begin () const noexcept
  {
    // TODO -- return const_iterator that points to the minimum node in the tree
    return const_iterator (nullptr);
  }

  const_iterator
  cbegin () const noexcept
  {
    return begin();
  }

  iterator
  end () noexcept
  {
    // TODO -- return iterator that points to the header node in the tree
    return iterator (nullptr);
  }

  const_iterator
  end () const noexcept
  {
    // TODO -- return const_iterator that points to the header node in the tree
    return const_iterator (nullptr);
  }

  const_iterator
  cend () const noexcept
  {
    return end();
  }

  reverse_iterator
  rbegin () noexcept
  {
    return reverse_iterator{end ()};
  }

  const_reverse_iterator
  rbegin () const noexcept
  {
    return const_reverse_iterator{end ()};
  }

  const_reverse_iterator
  crbegin () const noexcept
  {
    return const_reverse_iterator{cend ()};
  }

  reverse_iterator
  rend () noexcept
  {
    return reverse_iterator{begin ()};
  }

  const_reverse_iterator
  rend () const noexcept
  {
    return const_reverse_iterator{begin ()};
  }

  const_reverse_iterator
  crend () const noexcept
  {
    return const_reverse_iterator{cbegin ()};
  }

  bool
  empty () const noexcept
  {
    return size () == 0;
  }

  size_type
  size () const noexcept
  {
    return m_tree.size ();
  }

  size_type
  max_size () const noexcept
  {
    return std::numeric_limits<size_type>::max ();
  }

  void
  clear () noexcept
  {
    m_tree.clear ();
  }

  std::pair<iterator, bool>
  insert (value_type const& v)
  {
    auto [ptr, inserted] = m_tree.insert (v);
    return std::make_pair (iterator (ptr), inserted);
  }

  template<typename InputIterator,
           Requires (concepts::ForwardIterator, InputIterator)>
  void
  insert (InputIterator first, InputIterator last)
  {
    // TODO -- insert the range [first, last) into the set
  }

  void
  insert (std::initializer_list<T> ilist)
  {
    insert (ilist.begin (), ilist.end ());
  }

  iterator
  erase (const_iterator first, const_iterator last)
  {
    // TODO -- erase the range [first, last) from the set
    // hint: call erase in a loop. You MUST use the return value from erase()
    return iterator (nullptr);
  }

  iterator
  erase (iterator i)
  {
    node_type* n = m_tree.erase (i.m_node);
    if (n == nullptr)
    {
      return end ();
    }
    return iterator (n);
  }

  size_t
  erase (value_type const& v)
  {
    node_type* n = m_tree.find (v);
    if (n == nullptr)
    {
      return 0;
    }
    m_tree.erase (n);
    return 1;
  }

  constexpr void
  swap (Set& other) noexcept
  {
    using std::swap;
    swap (m_tree, other.m_tree);
  }

  constexpr iterator
  find (value_type const& value) const
  {
    node_type* ptr = m_tree.find (value);
    if (ptr == nullptr)
    {
      return end ();
    }
    return iterator (ptr);
  }

  constexpr size_type
  count (value_type const& value) const
  {
    return static_cast<size_type> (find (value) != end ());
  }

  constexpr bool
  contains (value_type const& value) const
  {
    return find (value) != end ();
  }

private:
  Tree<T> m_tree;
};

#endif
