#ifndef TREE_HPP_
#define TREE_HPP_

#include <utility>

#include "TreeNode.hpp"

template<typename T, bool AllowDuplicates = false>
struct Tree
{
  using Node = TreeNode<T>;

  constexpr Node*
  header () const noexcept
  {
    return const_cast<Node*> (&m_header);
  }

  constexpr Node*&
  root () const noexcept
  {
    return const_cast<Node*&> (m_header.parent);
  }

  constexpr Node*&
  min () const noexcept
  {
    return const_cast<Node*&> (m_header.right);
  }

  constexpr Node*&
  max () const noexcept
  {
    return const_cast<Node*&> (m_header.left);
  }

  // Repair the min and max as if node was removed from the tree
  // - if node is the minimum node, update to the successor
  // - if node is the maximum node, update to the predecessor
  constexpr void
  fixupExtremesErase (Node* node) noexcept
  {
    if (min () == node)
    {
      min () = TreeNodeUtils::successor (node);
    }
    if (max () == node)
    {
      max () = TreeNodeUtils::predecessor (node);
    }
  }

  // Update the min and/or max as if node was inserted into the tree
  // - if min is header or node's data is now the smallest value, update min to data
  // - if max is header or node's data is now the largest value, update max to data
  constexpr void
  fixupExtremesInsert (Node* node) noexcept
  {
    // TODO
  }

  // ------------------------------ PUBLIC API ------------------------------

  Tree () : m_header (), m_size (0)
  {
    root () = nullptr;
    max () = header ();
    min () = header ();
  }

  // copy constructor
  // - root should be a copy of other's root -- remember to use header as parent!
  // - max should be the maximum of the copied tree
  // - min should be the minimum of the copied tree
  Tree (Tree const& other) : m_header (), m_size (other.m_size)
  {
    // TODO
  }

  constexpr Tree (Tree&& other) noexcept
    : m_header (std::exchange (other.m_header, Node ()))
    , m_size (std::exchange (other.m_size, 0))
  {
    if (root () != nullptr) // fix root's parent to the header
    {
      root ()->parent = header ();
    }
  }

  // destructor
  //   should clear all elements from the tree
  ~Tree () noexcept
  {
    clear ();
  }

  // copy assignment
  //   should clear all elements from the tree and then copy all
  //   elements from other to "this" tree. Remember to update size!
  //   as always -- CHECK FOR SELF ASSIGNMENT
  Tree&
  operator= (Tree const& other)
  {
    // TODO
    return *this;
  }

  // move assignment
  constexpr Tree&
  operator= (Tree&& other) noexcept
  {
    // moves header and size -- fixes root()->parent
    TreeNodeUtils::clear (root());
    m_header = std::exchange (other.m_header, Node ());
    m_size = std::exchange (other.m_size, 0);
    if (root () != nullptr)
    {
      root ()->parent = header ();
    }
    return *this;
  }

  // returns the number of elements currently in the tree
  constexpr size_t
  size () const noexcept
  {
    return m_size;
  }

  // clears the tree
  // - sets min and max to dummy header
  // - removes the root (if any) and all its descendents
  // - sets root to nullptr
  // - sets the size to 0
  void
  clear () noexcept
  {
    // TODO
  }

  std::pair<Node*, bool>
  insert (T const& value)
  {
    return insert (value, root (), header ());
  }

  constexpr Node*
  find (T const& value) const noexcept
  {
    return find (value, root ());
  }

  std::pair<Node*, bool>
  insert (T&& value)
  {
    return insert (std::move (value), root (), header ());
  }

  // Remove a known node from the tree
  //
  // Case 0: node is nullptr
  // - return nullptr
  //
  // * Determine in-order successor of n -- "ios"
  // * get a reference to node -- "n"
  //
  // Case 1: node is a leaf
  // - delete "n"
  // - assign "n" to nullptr
  //
  // Case 2: node has only a left child
  // - save left child
  // - delete "n"
  // - assign "n" to saved left child
  //
  // Case 3: node has only a right child
  // - save right child
  // - delete "n"
  // - assign "n" to saved right child
  //
  // Case 4: node has left and right children
  // - Special Case: n's right child is not the in-order successor
  //   - replace ios with ios->right (remember, ios is a left child)
  //     and to update the parent link!
  //   - set ios->right to be n->right (and update parent link
  //     if ios->right isn't null)
  //   - continue with the steps below
  // - Splice in ios by taking over parent and left links
  //   e.g. ios->parent = n->parent   AND ios->left = n->left
  //   NOTE: remember to update the parent link as well (if non-null)
  // - null out n->left and n->right
  // - delete "n"
  // - assign "n" to "ios"
  //
  Node*
  erase (Node* node)
  {
    if (node == nullptr)
    {
      return nullptr;
    }

    Node* ios = TreeNodeUtils::successor (node);
    Node*& n = TreeNodeUtils::referenceOf (node);
    fixupExtremesErase (node);

    // TODO: if-else of cases goes here

    --m_size;
    return ios;
  }

private:
  // Find value function. Given a Node*, return it IFF it's data
  // is equal to value. Should return nullptr if the value isn't
  // found in the tree. Remember to "search" correctly
  constexpr Node*
  find (T const& value, Node* node) const noexcept
  {
    // TODO
    return nullptr;
  }

  // Recursive insertion routine
  // - value: value we are inserting
  // - node: reference to Node* (where we are currently trying to insert)
  // - parent: pointer to parent of node
  //
  // Note: we should always insert when node == nullptr
  // - remember to update size, assign the new Node to "node", and fixup extremes
  // I am providing partial code which is commented out below
  std::pair<Node*, bool>
  insert (T const& value, Node*& node, Node* parent)
  {
    if (node == nullptr)
    {
      // TODO
    }
    /*

      TODO: Add else-if cases here

    */
    if constexpr (AllowDuplicates)
    {
      return insert (value, node->left, node);
    }
    /* no duplicates allowed */
    return std::make_pair (node, false);
  }

  // Same as above but recursive calls should say std::move (value) instead
  // of "value". Construction of the new Node should also state std::move (value)
  std::pair<Node*, bool>
  insert (T&& value, Node*& node, Node* parent)
  {
    if (node == nullptr)
    {
      // TODO
    }
    /*

      TODO: Add else-if cases here

    */
    if constexpr (AllowDuplicates)
    {
      return insert (std::move (value), node->left, node);
    }
    /* no duplicates allowed */
    return std::make_pair (node, false);
  }

private:
  Node m_header;
  size_t m_size;
};

#endif
