#ifndef TREE_UTIL_HPP_
#define TREE_UTIL_HPP_

#include "Tree.hpp"

#include <ostream>
#include <sstream>

class PrintVisitor
{
  std::ostream& os;
  bool printed{false};

public:
  inline PrintVisitor (std::ostream& out) : os (out)
  {
  }

  inline void
  beforeVisit ()
  {
  }

  inline void
  afterVisit ()
  {
  }

  template<typename T>
  void
  visit (TreeNode<T>* node)
  {
    if (printed)
    {
      os << ' ';
    }
    else
    {
      printed = true;
    }
    os << node->data;
  }
};

class TreePrintVisitor
{
  std::ostream& os;
  int depth;
  int indent;

public:
  inline TreePrintVisitor (std::ostream& out, int indent)
    : os (out), depth (-indent), indent (indent)
  {
  }

  inline void
  beforeVisit ()
  {
    depth += indent;
  }

  inline void
  afterVisit ()
  {
    depth -= indent;
  }

  template<typename T>
  void
  visit (TreeNode<T>* node)
  {
    for (int i = 0; i < depth; ++i)
    {
      os << ' ';
    }
    os << node->data << '\n';
  }
};

struct Options
{
  constexpr static int reversed = 0b1000;
  constexpr static int inorder = 0b0001;
  constexpr static int preorder = 0b0010;
  constexpr static int postorder = 0b0100;
};

template<int Mask>
struct TreeVisitor
{
  static_assert ((Mask <= 0b1111) && (Mask > 0), "Invalid options");
  static_assert (!((Mask & ~Options::reversed) &
                   ((Mask & ~Options::reversed) - 1)),
                 "More than one ordering policy specified");

  template<typename T, typename Action>
  static void
  visit (TreeNode<T>* node, Action& action)
  {
    if (node == nullptr)
      return;
    action.beforeVisit ();
    if constexpr (Mask & Options::preorder)
      action.visit (node);
    if constexpr (Mask & Options::reversed)
      TreeVisitor::visit (node->right, action);
    else
      TreeVisitor::visit (node->left, action);
    if constexpr (Mask & Options::inorder)
      action.visit (node);
    if constexpr (Mask & Options::reversed)
      TreeVisitor::visit (node->left, action);
    else
      TreeVisitor::visit (node->right, action);
    if constexpr (Mask & Options::postorder)
      action.visit (node);
    action.afterVisit ();
  }
};

template<int Mask, typename T>
std::string
to_string (TreeNode<T>* root)
{
  std::ostringstream oss;
  PrintVisitor printer{oss};
  TreeVisitor<Mask>::visit (root, printer);
  return oss.str ();
}

template<typename T>
std::string
as_tree (TreeNode<T>* root)
{
  std::ostringstream oss;
  oss << '\n';
  TreePrintVisitor printer{oss, 2};
  TreeVisitor<Options::reversed | Options::inorder>::visit (root, printer);
  return oss.str ();
}

#endif
