#include "common.hpp"

#include "catch_config.hpp"

#include "catch.hpp"

#include "../Tree.hpp"

#include "../TreeUtils.hpp"



template <typename T>
size_t
compare (TreeNode<T>* actual, TreeNode<T>* actualParent, TreeNode<T>* expected)
{
  if (expected == nullptr) {
    REQUIRE (actual == nullptr);
    return 0;
  }
  {
    INFO ("data values match");
    REQUIRE (actual->data == expected->data);
  }
  {
    INFO ("parent pointer should match actual parent");
    REQUIRE (actual->parent == actualParent);
  }
  size_t size = 1;
  size += compare (actual->left, actual, expected->left);
  size += compare (actual->right, actual, expected->right);
  return size;
}

template <typename T>
void compare_trees (Tree<T>& actual, Tree<T>& expected) {
  REQUIRE (as_tree (actual.root()) == as_tree (expected.root()));
  size_t numNodes = compare (actual.root(), actual.header(), expected.root());
  REQUIRE (numNodes == actual.size());
  if (numNodes > 0)
  {
    REQUIRE (actual.min()->data == expected.min()->data);
    REQUIRE (actual.max()->data == expected.max()->data);
  }
}


SCENARIO ("fixupExtremesInsert updates min/max", "[Tree][fixup]")
{
  GIVEN ("an empty tree")
  {
    WHEN ("We call fixup with a node")
    {
      Tree<Counter<int>> t;
      TreeNode<Counter<int>> n;
      Counters::reset ();
      t.fixupExtremesInsert (&n);
      THEN ("[1] min and max are both updated; no comparisons are performed")
      {
        REQUIRE (t.min () == &n);
        REQUIRE (t.max () == &n);
        REQUIRE (Counters::compares == 0);
      }
      t.min () = t.max () = nullptr;
    }
  }
  GIVEN ("a tree with min and max set")
  {
    Tree<Counter<int>> t;
    TreeNode<Counter<int>> min;
    TreeNode<Counter<int>> max;
    min.data = Counter<int> (1);
    max.data = Counter<int> (5);
    t.min () = &min;
    t.max () = &max;

    TreeNode<Counter<int>> smallest (Counter<int> (0));
    TreeNode<Counter<int>> largest (Counter<int> (10));
    TreeNode<Counter<int>> middle (Counter<int> (4));

    WHEN ("We call fixup with a new smallest node")
    {
      Counters::reset ();
      t.fixupExtremesInsert (&smallest);
      THEN ("[1] min is updated. max stays the same; two comparisons performed")
      {
        REQUIRE (t.min () == &smallest);
        REQUIRE (t.max () != &smallest);
        REQUIRE (Counters::compares == 2);
      }
      t.min () = t.max () = nullptr;
    }
    WHEN ("We call fixup with a new largest node")
    {
      Counters::reset ();
      t.fixupExtremesInsert (&largest);
      THEN ("[1] min is updated. max stays the same; two comparisons performed")
      {
        REQUIRE (t.max () == &largest);
        REQUIRE (t.min () != &largest);
        REQUIRE (Counters::compares == 2);
      }
      t.min () = t.max () = nullptr;
    }
    WHEN ("We call fixup with a node between min and max")
    {
      Counters::reset ();
      t.fixupExtremesInsert (&middle);
      THEN ("[1] min and max stay the same; two comparisons performed")
      {
        REQUIRE (t.max () != &middle);
        REQUIRE (t.min () != &middle);
        REQUIRE (Counters::compares == 2);
      }
      t.min () = t.max () = nullptr;
    }
  }
}

SCENARIO ("Tree insert works", "[Tree][insert]")
{
  GIVEN ("An empty tree")
  {
    Tree<int> t;
    WHEN ("We insert a node into the tree")
    {
      int val = 4;
      auto returnValue = t.insert (val);
      THEN ("[1] the invariant of the tree is held")
      {
        {
          INFO ("The node should be returned")
          REQUIRE (returnValue.first == t.root ());
        }
        {
          INFO ("the second part of the pair must always be true");
          REQUIRE (returnValue.second);
        }
        {
          INFO ("Size should be increased");
          REQUIRE (t.size () == 1);
        }
        {
          INFO ("It was inserted into the correct location");
          REQUIRE (t.header ()->parent == t.root ());
          REQUIRE (t.root ()->parent == t.header ());
          REQUIRE (t.root ()->data == 4);
        }
        {
          INFO ("min and max should have been updated");
          REQUIRE (t.root () == t.min ());
          REQUIRE (t.root () == t.max ());
        }
      }
    }
    WHEN ("We r-value insert a node into the tree")
    {
      int val = 4;
      auto returnValue = t.insert (std::move(val));
      THEN ("[1] the invariant of the tree is held")
      {
        {
          INFO ("The node should be returned")
          REQUIRE (returnValue.first == t.root ());
        }
        {
          INFO ("the second part of the pair must always be true");
          REQUIRE (returnValue.second);
        }
        {
          INFO ("Size should be increased");
          REQUIRE (t.size () == 1);
        }
        {
          INFO ("It was inserted into the correct location");
          REQUIRE (t.header ()->parent == t.root ());
          REQUIRE (t.root ()->parent == t.header ());
          REQUIRE (t.root ()->data == 4);
        }
        {
          INFO ("min and max should have been updated");
          REQUIRE (t.root () == t.min ());
          REQUIRE (t.root () == t.max ());
        }
      }
    }
  }
  GIVEN ("A tree with a value")
  {
    Tree<int> t;
    TreeNode<int>* n = new TreeNode<int> (4);
    t.root () = t.min () = t.max () = n;
    n->parent = t.header ();
    size_t const OLD_SIZE = t.size ();
    WHEN ("We insert a node into the tree with the same value")
    {
      int val = 4;
      auto returnValue = t.insert (val);
      THEN ("[1] Nothing should change")
      {
        REQUIRE (std::make_pair (t.root (), false) == returnValue);
        REQUIRE (t.size () == OLD_SIZE);
        REQUIRE (t.header ()->parent == t.root ());
        REQUIRE (t.root ()->parent == t.header ());
        REQUIRE (t.root ()->data == 4);
        REQUIRE (t.root () == t.min ());
        REQUIRE (t.root () == t.max ());
      }
    }
    WHEN ("We insert a node into the tree which is the smallest")
    {
      int val = 2;
      auto returnValue = t.insert (val);
      THEN ("[1] the node was inserted and min was updated")
      {
        REQUIRE (std::make_pair (t.min (), true) == returnValue);
        REQUIRE (t.size () == OLD_SIZE + 1);
        REQUIRE (t.max () == t.root ());
        REQUIRE (t.min ()->parent == t.root ());
        REQUIRE (t.root ()->left == t.min ());
        REQUIRE (t.root ()->parent == t.header ());
      }
    }
    WHEN ("We insert a node into the tree which is the largest")
    {
      int val = 8;
      auto returnValue = t.insert (val);
      THEN ("[1] the node was inserted and max was updated")
      {
        REQUIRE (std::make_pair (t.max (), true) == returnValue);
        REQUIRE (t.size () == OLD_SIZE + 1);
        REQUIRE (t.min () == t.root ());
        REQUIRE (t.max ()->parent == t.root ());
        REQUIRE (t.root ()->right == t.max ());
        REQUIRE (t.root ()->parent == t.header ());
      }
      AND_WHEN ("We insert a node into the middle")
      {
        int val = 6;
        auto returnValue = t.insert (val);
        THEN ("[1] the structure of tree is correct; max/min don't change")
        {
          auto inserted = returnValue.first;

          REQUIRE (returnValue.second);
          REQUIRE (t.size () == OLD_SIZE + 2);
          REQUIRE (inserted->data == 6);
          REQUIRE (t.min ()->data == 4);
          REQUIRE (t.max ()->data == 8);
          REQUIRE (t.root () == t.min ());

          REQUIRE (t.root ()->parent == t.header ());

          REQUIRE (t.root ()->right == t.max ());
          REQUIRE (t.max ()->parent == t.root ());

          REQUIRE (t.max ()->left == inserted);
          REQUIRE (inserted->parent == t.max ());
        }
      }
    }


    WHEN ("We r-value insert a node into the tree with the same value")
    {
      int val = 4;
      auto returnValue = t.insert (std::move(val));
      THEN ("[1] Nothing should change")
      {
        REQUIRE (std::make_pair (t.root (), false) == returnValue);
        REQUIRE (t.size () == OLD_SIZE);
        REQUIRE (t.header ()->parent == t.root ());
        REQUIRE (t.root ()->parent == t.header ());
        REQUIRE (t.root ()->data == 4);
        REQUIRE (t.root () == t.min ());
        REQUIRE (t.root () == t.max ());
      }
    }
    WHEN ("We r-value insert a node into the tree which is the smallest")
    {
      int val = 2;
      auto returnValue = t.insert (std::move(val));
      THEN ("[1] the node was inserted and min was updated")
      {
        REQUIRE (std::make_pair (t.min (), true) == returnValue);
        REQUIRE (t.size () == OLD_SIZE + 1);
        REQUIRE (t.max () == t.root ());
        REQUIRE (t.min ()->parent == t.root ());
        REQUIRE (t.root ()->left == t.min ());
        REQUIRE (t.root ()->parent == t.header ());
      }
    }
    WHEN ("We r-value insert a node into the tree which is the largest")
    {
      int val = 8;
      auto returnValue = t.insert (std::move(val));
      THEN ("[1] the node was inserted and max was updated")
      {
        REQUIRE (std::make_pair (t.max (), true) == returnValue);
        REQUIRE (t.size () == OLD_SIZE + 1);
        REQUIRE (t.min () == t.root ());
        REQUIRE (t.max ()->parent == t.root ());
        REQUIRE (t.root ()->right == t.max ());
        REQUIRE (t.root ()->parent == t.header ());
      }
      AND_WHEN ("We r-value insert a node into the middle")
      {
        int val = 6;
        auto returnValue = t.insert (std::move(val));
        THEN ("[1] the structure of tree is correct; max/min don't change")
        {
          auto inserted = returnValue.first;

          REQUIRE (returnValue.second);
          REQUIRE (t.size () == OLD_SIZE + 2);
          REQUIRE (inserted->data == 6);
          REQUIRE (t.min ()->data == 4);
          REQUIRE (t.max ()->data == 8);
          REQUIRE (t.root () == t.min ());

          REQUIRE (t.root ()->parent == t.header ());

          REQUIRE (t.root ()->right == t.max ());
          REQUIRE (t.max ()->parent == t.root ());

          REQUIRE (t.max ()->left == inserted);
          REQUIRE (inserted->parent == t.max ());
        }
      }
    }
  }
}

SCENARIO ("Tree find works", "[Tree][find]")
{
  GIVEN ("An empty Tree")
  {
    Tree<int> t;
    WHEN ("We find a value")
    {
      auto result = t.find (5);
      THEN ("[0.25] we return nullptr")
      {
        REQUIRE (result == nullptr);
      }
    }
  }
  GIVEN ("A Tree with some values")
  {
    Tree<int> t;
    t.insert (16);
    t.insert (8);
    t.insert (24);
    t.insert (4);
    t.insert (12);
    t.insert (20);
    t.insert (28);
    WHEN ("We find a minimum value that doesn't exist")
    {
      auto result = t.find (1);
      THEN ("[0.25] we return nullptr")
      {
        REQUIRE (result == nullptr);
      }
    }
    AND_WHEN ("We find a value between two that doesn't exist")
    {
      auto result = t.find (15);
      THEN ("[0.25] we return nullptr")
      {
        REQUIRE (result == nullptr);
      }
    }
    AND_WHEN ("We find a maximum value that doesn't exist")
    {
      auto result = t.find (32);
      THEN ("[0.25] we return nullptr")
      {
        REQUIRE (result == nullptr);
      }
    }
    AND_WHEN ("We find() the smallest value in the tree")
    {
      auto result = t.find (4);
      THEN ("[1] we return a pointer to min()")
      {
        REQUIRE (result == t.min ());
        REQUIRE (result != nullptr);
        REQUIRE (result->data == 4);
      }
    }
    AND_WHEN ("We find() the largest value in the tree")
    {
      auto result = t.find (28);
      THEN ("[1] we return a pointer to max()")
      {
        REQUIRE (result == t.max ());
        REQUIRE (result != nullptr);
        REQUIRE (result->data == 28);
      }
    }
    AND_WHEN ("We find() some other value in the tree")
    {
      auto result = t.find (12);
      THEN ("[1] we return a pointer to that node")
      {
        REQUIRE (result != nullptr);
        REQUIRE (result->data == 12);
      }
    }
  }
}

SCENARIO ("Tree copy constructor works", "[Tree][cctor]")
{
  using T = Counter<int>;
  GIVEN ("An empty tree")
  {
    Tree<T> tree;
    WHEN ("We copy-construct the tree")
    {
      Counters::reset ();
      Tree<T> copy (tree);
      THEN ("[1] we made a copy")
      {
        REQUIRE (tree.min () != copy.min ());
        REQUIRE (tree.max () != copy.max ());
        REQUIRE (copy.root () == nullptr);
        REQUIRE (tree.size () == copy.size ());
        INFO ("we should construct exactly one Node (for the header of copy)");
        REQUIRE (Counters::ctor == 1);
      }
    }
  }

  GIVEN ("A tree with some nodes")
  {
    T value;
    Tree<T> tree;
    for (int i : {4, 2, 6, 1, 3, 5, 7})
    {
      value.data = i;
      tree.insert (value);
    }
    WHEN ("We copy-construct the tree")
    {
      Counters::reset ();
      Tree<T> copy (tree);
      THEN ("[3] we made a copy")
      {
        REQUIRE (tree.root() != nullptr);
        {
          INFO ("We should construct exactly size() + 1 Nodes");
          REQUIRE (Counters::ctor == 1);
          REQUIRE (Counters::cctor == tree.size ());
        }
        {
          INFO ("Nodes references should be unique to each tree");
          REQUIRE (tree.min () != copy.min ());
          REQUIRE (tree.max () != copy.max ());
          REQUIRE (tree.root () != copy.root ());
        }
        {
          INFO ("Node elements should be copied");
          REQUIRE (tree.min ()->data == copy.min ()->data);
          REQUIRE (tree.max ()->data == copy.max ()->data);
          REQUIRE (tree.root ()->data == copy.root ()->data);
        }
        {
          INFO ("the trees should have the same size and shape");
          REQUIRE (tree.size () == copy.size ());
          REQUIRE (as_tree (copy.root ()) == as_tree (tree.root ()));
        }
      }
    }
  }
}

SCENARIO ("Tree clear works", "[Tree][clear]")
{
  using T = Counter<int>;
  GIVEN ("An empty tree")
  {
    Tree<T> tree;
    WHEN ("We clear the tree")
    {
      Counters::reset ();
      tree.clear ();
      THEN ("[1] Nothing should happen")
      {
        REQUIRE (Counters::dtor == 0);
        REQUIRE (tree.size () == 0);
        REQUIRE (tree.min () == tree.header ());
        REQUIRE (tree.max () == tree.header ());
        REQUIRE (tree.root () == nullptr);
      }
    }
  }
  GIVEN ("A tree with some elements")
  {
    Tree<T> tree;
    T value;
    for (int i : {4, 2, 6, 1, 3, 5, 7})
    {
      value.data = i;
      tree.insert (value);
    }
    WHEN ("We clear the tree")
    {
      Counters::reset ();
      tree.clear ();
      THEN ("[1] The tree is cleared/reset")
      {
        INFO ("destructor should be called size() times");
        REQUIRE (Counters::dtor == 7);
        INFO ("size should be zero");
        REQUIRE (tree.size () == 0);
        INFO ("min and max should refer to the header");
        REQUIRE (tree.min () == tree.header ());
        REQUIRE (tree.max () == tree.header ());
        INFO ("root should be nullptr");
        REQUIRE (tree.root () == nullptr);
      }
    }
  }
}

SCENARIO ("Tree copy assignment works", "[Tree][copy-assign]")
{
  using T = Counter<int>;
  GIVEN ("A tree")
  {
    Tree<T> lhs;
    Tree<T> rhs;
    T value;
    for (int i : {8, 4, 12, 2, 6, 10, 14, 1, 3, 5, 7, 9, 11, 13, 15})
    {
      value.data = i;
      lhs.insert (value);
    }
    WHEN ("We perform self assignment")
    {
      Counters::reset ();
      lhs = lhs;
      THEN ("[1] Nothing should happen")
      {
        REQUIRE (Counters::all_zero ("you should check for self-assignment"));
      }
    }
    WHEN ("We assign the lhs to an empty rhs")
    {
      Counters::reset ();
      lhs = rhs;
      THEN ("[1] we effectively clear")
      {
        REQUIRE (Counters::cctor == 0);
        REQUIRE (Counters::dtor >= 15);
        REQUIRE (lhs.size () == rhs.size ());
        REQUIRE (as_tree (lhs.root ()) == as_tree (rhs.root ()));
      }
    }
    WHEN ("We assign the lhs to a non-empty rhs")
    {
      for (int i : {2, 1, 3})
      {
        value.data = i;
        rhs.insert (value);
      }
      Counters::reset ();
      lhs = rhs;
      THEN ("[2] we effectively clear then copy")
      {
        REQUIRE (Counters::cctor == rhs.size ());
        REQUIRE (Counters::dtor >= 15);
        REQUIRE (lhs.size () == rhs.size ());
        REQUIRE (as_tree (lhs.root ()) == as_tree (rhs.root ()));
      }
    }
  }
}

template<typename T>
Tree<Counter<T>>
make_tree (std::initializer_list<T> ilist)
{
  Tree<Counter<T>> tree;
  Counter<T> value;
  for (T i : ilist)
  {
    value.data = i;
    tree.insert (value);
  }
  return tree;
}

SCENARIO ("Tree erase works", "[Tree][erase]")
{
  GIVEN ("A tree")
  {
    using T = Counter<int>;
    auto tree = make_tree ({6, 1, 10, 3, 8, 2, 4, 7, 9});
    /*
              6
             / \
            /   \
           /     \
          /       \
         1         10
          \       /
           3     8
          / \   / \
         2   4 7   9
     */

    auto const old_string = as_tree (tree.root ());
    auto const old_size = tree.size ();

    WHEN ("We try to erase a node that isn't in the tree (nullptr)")
    {
      T value (5);
      TreeNode<T>* node = tree.find (value);
      Counters::reset ();
      TreeNode<T>* returnVal = tree.erase (node);
      THEN ("[1] Nothing should change")
      {
        REQUIRE (
          Counters::all_zero ("shouldn't read/write/copy/destroy/create"));
        REQUIRE (old_string == as_tree (tree.root ()));
        REQUIRE (returnVal == nullptr);
        REQUIRE (old_size == tree.size ());
      }
    }

    WHEN ("We try to erase a node that is a leaf")
    {
      T value (2);
      TreeNode<T>* node = tree.find (value);
      REQUIRE (node != nullptr);
      TreeNode<T>*& parent = TreeNodeUtils::referenceOf (node->parent);
      Counters::reset ();
      TreeNode<T>* returnVal = tree.erase (node);
      THEN ("[2] the node should be removed")
      {
        REQUIRE (Counters::dtor == 1);
        REQUIRE (Counters::compares == 0);
        REQUIRE (old_size - 1 == tree.size ());
        REQUIRE (returnVal->data == T (3));
        REQUIRE (returnVal == parent);
        REQUIRE (parent->left == nullptr);
        /*
              6
             / \
            /   \
           /     \
          /       \
         1         10
          \       /
           3     8
            \   / \
             4 7   9
        */

        auto expected = make_tree ({6, 1, 10, 3, 8, 4, 7, 9});
        compare_trees (tree, expected);
      }
    }

    WHEN ("We try to erase a node that has only a left child")
    {
      T value (10);
      TreeNode<T>* node = tree.find (value);
      REQUIRE (node != nullptr);
      TreeNode<T>* successor = TreeNodeUtils::successor (node);
      Counters::reset ();
      TreeNode<T>* returnVal = tree.erase (node);
      THEN ("[2] the node should be removed")
      {
        REQUIRE (Counters::vctor + Counters::cctor + Counters::mctor + Counters::ctor + Counters::cassign + Counters::massign == 0);
        REQUIRE (Counters::dtor == 1);
        REQUIRE (Counters::compares == 0);
        REQUIRE (old_size - 1 == tree.size ());
        REQUIRE (returnVal == successor);
        auto expected = make_tree ({6, 1, 8, 3, 7, 9, 2, 4});
        compare_trees (tree, expected);
        /*
              6
             / \
            /   \
           /     \
          /       \
         1         8
          \       / \
           3     7   9
          / \
         2   4

        */
      }
    }
    WHEN ("We try to erase a node that has only a right child")
    {
      T value (1);
      TreeNode<T>* node = tree.find (value);
      REQUIRE (node != nullptr);
      TreeNode<T>* successor = TreeNodeUtils::successor (node);
      Counters::reset ();
      TreeNode<T>* returnVal = tree.erase (node);
      THEN ("[2] the node should be removed")
      {
        REQUIRE (Counters::vctor + Counters::cctor + Counters::mctor + Counters::ctor + Counters::cassign + Counters::massign == 0);
        REQUIRE (Counters::dtor == 1);
        REQUIRE (Counters::compares == 0);
        REQUIRE (old_size - 1 == tree.size ());
        REQUIRE (returnVal == successor);
        auto expected = make_tree ({6, 3, 10, 2, 4, 8, 7, 9});
        compare_trees (tree, expected);
        /*
              6
             / \
            /   \
           /     \
          /       \
         3         10
        / \       /
       2   4     8
                / \
               7   9
        */
      }
    }
    WHEN ("We try to remove a node that has two children")
    {
      WHEN ("Its in-order successor is its right child")
      {
        T value (8);
        TreeNode<T>* node = tree.find (value);
        REQUIRE (node != nullptr);
        TreeNode<T>* successor = TreeNodeUtils::successor (node);
        Counters::reset ();
        TreeNode<T>* returnVal = tree.erase (node);
        THEN ("[2] The node should be removed")
        {
          REQUIRE (Counters::vctor + Counters::cctor + Counters::mctor + Counters::ctor + Counters::cassign + Counters::massign == 0);
          REQUIRE (Counters::dtor == 1);
          REQUIRE (Counters::compares == 0);
          REQUIRE (old_size - 1 == tree.size ());
          REQUIRE (returnVal == successor);
          auto expected = make_tree ({6, 1, 10, 3, 9, 2, 4, 7});
          compare_trees (tree, expected);
        }
      }
      WHEN ("Its in-order successor is not its right child")
      {
        T value (6);
        TreeNode<T>* node = tree.find (value);
        REQUIRE (node != nullptr);
        TreeNode<T>* successor = TreeNodeUtils::successor (node);
        Counters::reset ();
        TreeNode<T>* returnVal = tree.erase (node);
        THEN ("[3] The node should be removed")
        {
          REQUIRE (Counters::vctor + Counters::cctor + Counters::mctor + Counters::ctor + Counters::cassign + Counters::massign == 0);
          REQUIRE (Counters::dtor == 1);
          REQUIRE (Counters::compares == 0);
          REQUIRE (old_size - 1 == tree.size ());
          REQUIRE (returnVal == successor);
          auto expected = make_tree ({7, 1, 10, 3, 8, 2, 4, 9});
          compare_trees (tree, expected);
        }
      }
    }
  }
}
